#
# MIT License
#
# (C) Copyright 2024 Hewlett Packard Enterprise Development LP
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
"""Tools for installing and manipulating vTDS layers.

"""

from os import makedirs
from os.path import join as path_join
import importlib
from .errors import ContextualError
from .config_operations import merge_configs


class Layer:
    """The class representation of a vTDS layer, used for calling that
    layer's API and obtaining that layer's base and test configuration
    data.

    """
    def __init__(self, module_name, is_core=False):
        """Constructor. The 'module_name' is used to import the layer
        and install the layer's API and base configuration implementation.

        """
        self.is_core = is_core
        try:
            module = importlib.import_module(module_name)
            self.layer_api_class = module.LayerAPI if not is_core else None
            self.base_config = module.BaseConfig()
            self.layer_api = None
        except ImportError as err:
            raise ContextualError(
                "cannot import layer module '%s' - %s" % (
                    module_name, str(err)
                )
            ) from err
        except AttributeError as err:
            raise ContextualError(
                "layer module '%s' does not implement a vTDS layer - %s" % (
                    module_name, str(err)
                )
            ) from err

    def initialize(self, stack, config, build_dir):
        """Initialize a layer, setting up its actual implementation
        based on the stack of layers loaded, the overall fully resolved
        vTDS configuration and the path to the build directory for the
        specific layer.

        """
        if self.is_core:
            return
        self.layer_api = self.layer_api_class(stack, config, build_dir)

    def get_api(self):
        """Return the layer API implementation object for use by the
        caller.

        """
        return self.layer_api


def __construct_layer__(module_name, is_core=False):
    """Given a module name that could be None or empty, return a Layer
    based on that module name or None if the name was None or empty.

    """
    return Layer(module_name, is_core) if module_name else None


class VTDSStack:
    """The class representation of the full stack of layers:
    application, cluster, platform and provider.

    """
    def __init__(
            self,
            application_name=None,
            cluster_name=None,
            platform_name=None,
            provider_name=None,
            core_name=None
    ):
        """Constructor. Create a stack using the Python module names
        for the application, cluster, platform and provider layers
        specified in the 'application_name', 'cluster_name',
        'platform_name' and 'provider_name' arguments respectively.

        """
        core_name = "vtds_core" if not core_name else core_name
        self.application = __construct_layer__(application_name)
        self.cluster = __construct_layer__(cluster_name)
        self.platform = __construct_layer__(platform_name)
        self.provider = __construct_layer__(provider_name)
        self.core = __construct_layer__(core_name, is_core=True)
        self.config = None

    def __init_layer__(self, name, layer, config, build_dir):
        """Initialize a layer where 'name' identifies the name of the
        layer in the stack (used to sub-divide the build directory
        tree), 'layer' is the Layer object containing the layer
        information, 'config' is the full vTDS configuration to be
        supplied the layer and 'build_dir' is the root of the build
        directory tree. The build directory tree and sub-directories
        for each layer are created as needed.

        """
        if layer is None:
            # No such layer present, this is not an error, but we
            # don't want to go any further.
            return
        layer_build_dir = path_join(build_dir, name)
        try:
            makedirs(layer_build_dir, 0o700, True)
        except OSError as err:
            raise ContextualError(
                "failed to create build directory for layer "
                "'%s' ['%s'] - %s" % (name, layer_build_dir, str(err))
            ) from err
        layer.initialize(self, config, layer_build_dir)

    def __active_apis__(self, api_list=None):
        """Get the list of active APIs in the stack. If 'api_list' is
        provided, then get the APIs in the order specified. Otherwise,
        get them in the following order: 'provider', 'platform',
        'cluster', 'application'.

        """
        api_list = (
            [
                self.provider.get_api() if self.provider else None,
                self.platform.get_api() if self.platform else None,
                self.cluster.get_api() if self.cluster else None,
                self.application.get_api() if self.application else None,
            ] if api_list is None else api_list
        )
        return [api for api in api_list if api is not None]

    def __active_configs__(self, config_list=None):
        """Get the list of active base configs in the stack. If
        'config_list' is provided, then get the configs in the order
        specified. Otherwise, get them in the following order:
        'provider', 'platform', 'cluster', 'application'.

        """
        config_list = (
            [
                self.provider.base_config if self.provider else None,
                self.platform.base_config if self.platform else None,
                self.cluster.base_config if self.cluster else None,
                self.application.base_config if self.application else None,
                self.core.base_config if self.core else None,
            ] if config_list is None else config_list
        )
        return [config for config in config_list if config is not None]

    def initialize(self, config, build_dir):
        """Initialize all of the layers rooted at 'build_dir' and
        supplying the full vTDS configuration found in 'config'. Each
        layer will have a sub-directory created for it in the build
        directory tree.

        """
        self.__init_layer__("core", self.core, config, build_dir)
        self.__init_layer__("application", self.application, config, build_dir)
        self.__init_layer__("cluster", self.cluster, config, build_dir)
        self.__init_layer__("platform", self.platform, config, build_dir)
        self.__init_layer__("provider", self.provider, config, build_dir)
        self.config = config

    def consolidate(self):
        """Execute the stack 'consolidate' phase. This runs the
        'consolidate()' method in each of the layers from the bottom
        to the top to set up for any phases that rely on a fully
        resolved configuration.

        """
        for api in self.__active_apis__():
            api.consolidate()

    def prepare(self):
        """Execute the stack 'prepare' phase. This runs the
        'prepare()' method in each of the layers from the bottom to
        the top to set up for the 'validate' and 'deploy' phases.

        """
        for api in self.__active_apis__():
            api.prepare()

    def validate(self):
        """Execute the stack 'validate' phase. This runs the
        'validate()' method in each of the layers from the bottom to
        the top to ensure that the layers are ready for the 'deploy'
        phase.

        """
        for api in self.__active_apis__():
            api.validate()

    def deploy(self):
        """Execute the stack 'deploy' phase. This runs the
        'deploy()' method in each of the layers from the bottom to
        the top.

        """
        for api in self.__active_apis__():
            api.deploy()

    def remove(self):
        """Execute the stack 'remove' phase. This runs the
        'remove()' method in each of the layers from the top to
        the bottom to clean up all resources used by the vTDS.

        """
        # Work from top to bottom instead of bottom to top to avoid
        # pulling a dependency out from under an upper layer.
        api_list = [
                self.application.get_api() if self.application else None,
                self.cluster.get_api() if self.cluster else None,
                self.platform.get_api() if self.platform else None,
                self.provider.get_api() if self.provider else None,
        ]
        for api in self.__active_apis__(api_list):
            api.remove()

    def get_base_config_text(self):
        """Collate the annotated base configurations of all of the
        layers that are present into a single YAML string and return
        it to the caller.

        """
        ret = ""
        configs = self.__active_configs__()
        for config in configs:
            ret += config.get_base_config_text() + '\n'
        return ret

    def get_base_config(self):
        """Collate and merge the base configurations of all of the
        layers that are present into a single base configuration and
        return it to the caller as a configuration collection for use
        in constructing a config to deploy a vTDS.

        """
        ret = {}
        configs = self.__active_configs__()
        for config in configs:
            ret = merge_configs(ret, config.get_base_config())
        return ret

    def get_final_config(self):
        """Return the full configuration after all of the overlays
        have been applied that is given to the layers. This only works
        after initialize() and consolidate() have been run...

        """
        if self.config is None:
            raise ContextualError(
                "can't request stack configuration before calling initialize()"
            )
        return self.config

    def get_test_config(self):
        """Get a base configuration and then apply all of the layer
        test overlays to it to get a test configuration.

        """
        base_config = self.get_base_config()
        configs = self.__active_configs__()
        for config in configs:
            base_config = merge_configs(base_config, config.get_test_overlay())
        return base_config

    def get_application_api(self):
        """Accessor for the application layer API.

        """
        return (
            self.application.get_api() if self.application is not None
            else None
        )

    def get_application_base_config(self):
        """Accessor for the application layer base configuration..

        """
        return (
            self.application.base_config if self.application is not None
            else None
        )

    def get_cluster_api(self):
        """Accessor for the cluster layer API.

        """
        return (
            self.cluster.get_api() if self.cluster is not None
            else None
        )

    def get_cluster_base_config(self):
        """Accessor for the cluster layer base configuration..

        """
        return (
            self.cluster.base_config if self.cluster is not None
            else None
        )

    def get_platform_api(self):
        """Accessor for the platform layer API.

        """
        return (
            self.platform.get_api() if self.platform is not None
            else None
        )

    def get_platform_base_config(self):
        """Accessor for the platform layer base configuration..

        """
        return (
            self.platform.base_config if self.platform is not None
            else None
        )

    def get_provider_api(self):
        """Accessor for the provider layer API.

        """
        return (
            self.provider.get_api() if self.provider is not None
            else None
        )

    def get_provider_base_config(self):
        """Accessor for the provider layer base configuration..

        """
        return (
            self.provider.base_config if self.provider is not None
            else None
        )



#
# MIT License
#
# (C) Copyright 2024 Hewlett Packard Enterprise Development LP
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
"""Functions for executing commands derived from the subprocess module.

"""
import subprocess

from .errors import ContextualError
from .logs import logfile


def run(cmd, log_files=None, **run_args):
    """Run a command. All of the standard subprocess.run() arguments
    are honored from the caller if they are provided. Returns the
    CompletedProcess object returned from the subprocess.run() call
    which contains, the return code from the command, and any stdout
    or stderr collected from the command.  By default, stdout and
    stderr are collected for logging purposes and the 'check' argument
    is set to 'True' which causes __run() to raise a ContextualError
    if the command fails. To capture 'stdout' and / or 'stderr' in the
    CompletedProcess instead, set either or both to
    'subprocess.PIPE'. To capture a non-zero exit value, set 'check'
    to False.

    The 'log_files' argument contains a tuple of either pathnames to
    files or open file streams for use in capturing output on stdout
    and stderr respectively. If either is None, output on that stream
    is discarded.  The behavior of 'log_files' can be explicitly
    overridden by setting the 'stdout' and 'stderr' keyword arguments
    to run() in the call.

    """
    log_files = (
        log_files
        if log_files is not None else
        ("/dev/null", "/dev/null")
    )
    out_file, err_file = log_files
    with logfile(out_file) as out, logfile(err_file) as err:
        try:
            # We want to set certain default arguments to run() if
            # they aren't overridden by the caller.
            run_actual = {
                'stdout': out,
                'stderr': err,
                'text': True,
                'encoding': 'UTF-8',
            }
            run_actual.update(run_args)
            check = run_actual.pop('check', True)
            completion = subprocess.run(cmd, check=check, **run_actual)
        except subprocess.CalledProcessError as err:
            raise ContextualError(
                "execution of '%s' command failed - %s" % (
                    " ".join(cmd),
                    str(err)
                ),
                out_file if isinstance(out_file, str) else None,
                err_file if isinstance(err_file, str) else None
            ) from err
    return completion



#
# MIT License
#
# (C) Copyright 2024-2025 Hewlett Packard Enterprise Development LP
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
"""Objects presented on the Layer API containing public information
and operations in the provider layer.

"""
from abc import (
    ABCMeta,
    abstractmethod
)


class SiteConfigBase(metaclass=ABCMeta):
    """A class implementing public access to site configuration
    settings.

    """

    @abstractmethod
    def system_name(self):
        """Get the system name used to create the cluster on the
        provider. Returns a string representing the system name.

        """

    @abstractmethod
    def site_ntp_servers(self, address_family='AF_INET'):
        """Get the hostname / address of the NTP server offered by the
        provider or site in the specified address family. The
        'address_family' parameter is a string specifies the address
        family of the address to be returned, with a default value of
        'AF_INET'. Returns a list of dictionaries containing the
        following fields:

        - 'address_family' - the requested address family as passed in
        - 'hostname' - the host name of the NTP server if known, None if not
        - 'address' - the address within the address family of the NTP
          server if known, None if not

        Returns an empty list if the specified 'address_family' is not
        supported by the provider or not configured in the Provider
        configuration.

        """

    @abstractmethod
    def site_dns_servers(self, address_family='AF_INET'):
        """Get the address of the DNS server offered by the provider
        or site on the specified address family. The address_family'
        parameter is a string specifies the address family of the
        address to be returned, with a default value of
        'AF_INET'. Returns a list of dictionarie containing the
        following fields:

        - 'address_family' - the requested address family as passed in
        - 'address' - the address within the address family of the NTP
          server if known, None if not

        Returns an empty list if the specified 'address_family' is not
        supported by the provider or not configured in the Provider
        configuration.

        """


class VirtualBladesBase(metaclass=ABCMeta):
    """A class implementing public access to Virtual Blades and their
    operations.

    """
    @abstractmethod
    def blade_classes(self):
        """Get a list of virtual blade classes by name.

        """

    @abstractmethod
    def application_metadata(self, blade_class):
        """Get the application metadata for a named blade class from
        its config.

        """

    @abstractmethod
    def blade_count(self, blade_class):
        """Get the number of Virtual Blade instances of the specified
        class.

        """

    @abstractmethod
    def blade_interconnects(self, blade_class):
        """Return the list of Blade Interconnects by name connected to
        the specified class of Virtual Blade.

        """

    @abstractmethod
    def blade_hostname(self, blade_class, instance):
        """Get the hostname of a given instance of the specified class
        of Virtual Blade.

        """

    @abstractmethod
    def blade_ip(self, blade_class, instance, interconnect):
        """Return the IP address (string) on the named Blade
        Interconnect of a specified instance of the named Virtual
        Blade class.

        """

    @abstractmethod
    def blade_ssh_key_secret(self, blade_class):
        """Return the name of the secret containing the SSH key pair
        used to to authenticate with blades of the specified blade
        class.

        """

    @abstractmethod
    def blade_ssh_key_paths(self, blade_class):
        """Return a tuple of paths to files containing the public and
        private SSH keys used to to authenticate with blades of the
        specified blade class. The tuple is in the form '(public_path,
        private_path)' The value of 'private_path' is suitable for use
        with the '-i' option of 'ssh'. Before returning this call will
        verify that both files can be opened for reading and will fail
        with a ContextualError if either cannot.

        """

    @abstractmethod
    def connect_blade(self, blade_class, instance, remote_port):
        """Establish an external connection to the specified remote
        port on the specified instance of the named Virtual Blade
        class. Return a BladeConnection object for the connection.

        BladeConnections are context managed, so if this is used
        in a 'with' clause, upon leaving the 'with' context, the
        connection in the BladeConnection is automatically closed.

        To close the connection outside of a 'with' clause call
        the __exit__() method on the BladeConnection.

        """

    @abstractmethod
    def connect_blades(self, remote_port, blade_classes=None):
        """Establish external connections to the specified remote port
        on all the Virtual Blade instances on all the Virtual Blade
        classes listed by name in 'blade_classes'. If 'blade_classes' is not
        provided or None, all available blade classes are used. Return a
        BladeConnectionSet object representing the connections.

        BladeConnectionSets are context managed, so if this is used
        in a 'with' clause, upon leaving the 'with' context, the
        connections in the BladeConnectionSet are automatically closed.

        To close the connections outside of a 'with' clause call
        the __exit__() method on the BladeConnectionSet.

        """

    @abstractmethod
    def ssh_connect_blade(self, blade_class, instance, remote_port):
        """Establish an external connection to the SSH server
        port on the specified instance of the named Virtual Blade
        class. Return a BladeSSHConnection object for the
        connection.

        BladeSSHConnection are context managed, so if this is used
        in a 'with' clause, upon leaving the 'with' context, the
        connection in the BladeSSHConnection is automatically closed.

        To close the connection outside of a 'with' clause call
        the __exit__() method on the BladeSSHConnection.

        """

    @abstractmethod
    def ssh_connect_blades(self, blade_classes=None, remote_port=22):
        """Establish external connections to the specified remote port
        on all the Virtual Blade instances on all the Virtual Blade
        classes listed by name in 'blade_classes'. If 'blade_classes' is not
        provided or None, all available blade classes are used. Return a
        BladeSSHConnectionSet object representing the
        connections.

        BladeSSHConnectionSets are context managed, so if this is used
        in a 'with' clause, upon leaving the 'with' context, the
        connections in the BladeSSHConnectionSet are automatically closed.

        To close the connections outside of a 'with' clause call
        the __exit__() method on the BladeSSHConnectionSet.

        """


class BladeInterconnectsBase(metaclass=ABCMeta):
    """The external representation of the set of Blade Interconnects
    and public operations that can be performed on the interconnects.

    """
    @abstractmethod
    def interconnect_names(self):
        """Get a list of blade interconnects by name

        """

    @abstractmethod
    def application_metadata(self, interconnect_name):
        """Get the application metadata for a named interconnect from
        its config.

        """

    @abstractmethod
    def ipv4_cidr(self, interconnect_name):
        """Return the (string) IPv4 CIDR (<IP>/<length>) for the
        network on the named interconnect.

        """


class BladeConnectionBase(metaclass=ABCMeta):
    """A class containing the relevant information needed to use
    external connections to ports on a specific Virtual Blade.

    """
    @abstractmethod
    def blade_class(self):
        """Return the name of the Virtual Blade class of the connected
        Virtual Blade.

        """

    @abstractmethod
    def blade_hostname(self):
        """Return the hostname of the connected Virtual Blade.

        """

    @abstractmethod
    def local_ip(self):
        """Return the locally reachable IP address of the connection
        to the Virtual Blade.

        """

    @abstractmethod
    def local_port(self):
        """Return the TCP port number on the locally reachable IP
        address of the connection to the Virtual Blade.

        """

    @abstractmethod
    def remote_port(self):
        """Return the TCP port number on the on the Virtual blade to
        which the BladeConnection connects.

        """

    @abstractmethod
    def __enter__(self):
        """Context entry handler to make BladeConnection objects
        usable with the 'with ... as' construct. This returns the
        BladeConnection for use in a context.

        """

    @abstractmethod
    def __exit__(
            self,
            exception_type=None,
            exception_value=None,
            traceback=None
    ):
        """Context exit handler to make BladeConnection
        objects usable with the 'with ... as' construct. This cleans
        up all resources associated with the BladeConnection on
        exit from the 'with' block context. Not called if the object
        is used outside a 'with' context.

        """


class BladeConnectionSetBase(metaclass=ABCMeta):
    """A class that contains multiple active BladeConnections to
    facilitate operations on multiple simultaneous blades. This class
    is just a wrapper for a list of BladeContainers and should be
    obtained using the VirtualBlades.connect_blades() method not
    directly.

    """
    @abstractmethod
    def list_connections(self, blade_class=None):
        """List the connections in the BladeConnectionSet filtered by
        'blade_class' if that is present. Otherwise imply list all of
        the connections.

        """

    @abstractmethod
    def get_connection(self, hostname):
        """Return the connection corresponding to the specified
        VirtualBlade hostname ('hostname') or None if the hostname is
        not found.

        """

    @abstractmethod
    def __enter__(self):
        """Context entry handler to make BladeConnectionSet objects
        usable with the 'with ... as' construct. This returns the
        BladeConnectionSet for use in a context.

        """

    @abstractmethod
    def __exit__(
            self,
            exception_type=None,
            exception_value=None,
            traceback=None
    ):
        """Context exit handler to make BladeConnectionSet
        objects usable with the 'with ... as' construct. This cleans
        up all resources associated with the BladeConnectionSet on
        exit from the 'with' block context. Not called if the object
        is used outside a 'with' context.

        """


class BladeSSHConnectionBase(BladeConnectionBase, metaclass=ABCMeta):
    """Specifically a connection to the SSH server on a blade (remote
    port 22 unless otherwise specified) with methods to copy files to
    and from the blade using SCP and to run commands on the blade
    using SSH.

    """
    @abstractmethod
    def copy_to(
        self, source, destination,
        recurse=False, blocking=True, logname=None, **kwargs
    ):
        """Copy a file from a path on the local machine ('source') to
        a path on the virtual blade ('dest'). The SCP operation is run
        under a subprocess.Popen() object, which is returned at the
        end of the call. If additional keyword arguments are supplied,
        they may be used to override defaults set up by this function
        and passed to subprocess.Popen() or simply passed on to
        subprocess.Popen() as keyword arguments.

        If the 'recurse' argument is 'True' and the source is a
        directory, the directory and all of its descendents will be
        copied. Otherwise, the source should be a file and it alone
        will be copied.

        If the 'blocking' option is True (default), copy_to() will block
        waiting for the copy to complete (or fail) and raise a
        ContextualError exception if it fails. If the 'blocking' option
        is False, copy_to() will return immediately once the Popen()
        object is created and let the caller manage the sub-process.

        If the 'logname' argument is provided and not None, use the
        string found there to compose a pair of log files to capture
        standard output and standard error. Otherwise a generic log
        name is created.

        """

    @abstractmethod
    def copy_from(
        self, source, destination,
        recurse=False, blocking=True, logname=None, **kwargs
    ):
        """Copy a file from a path on the blade ('source') to a path
        on the local machine ('dest'). The SCP operation is run under
        a subprocess.Popen() object, which is returned at the end of
        the call. If additional keyword arguments are supplied, they
        may be used to override defaults set up by this function and
        passed to subprocess.Popen() or simply passed on to
        subprocess.Popen() as keyword arguments.

        If the 'recurse' argument is 'True' and the source is a
        directory, the directory and all of its descendents will be
        copied. Otherwise, the source should be a file and it alone
        will be copied.

        If the 'blocking' option is True (default), copy_from() will
        block waiting for the copy to complete (or fail) and raise a
        ContextualError exception if it fails. If the 'blocking' option
        is False, copy_from() will return immediately once the Popen()
        object is created and let the caller manage the sub-process.

        If the 'logname' argument is provided and not None, use the
        string found there to compose a pair of log files to capture
        standard output and standard error. Otherwise a generic log
        name is created.

        """

    @abstractmethod
    def run_command(self, cmd, blocking=True, logfiles=None, **kwargs):
        """Using SSH, run the command in the string 'cmd' on the
        blade. The string 'cmd' can be templated using Jinja
        templating to use any of the attributes of the underlying
        connection:

        - the blade class: 'blade_class'
        - the blade class instance: 'instance'
        - the blade hostname: 'blade_hostname'
        - the connection port on the blade: 'remote_port'
        - the local connection IP address: 'local_ip'
        - the local connection port: 'local_port'

        The resulting command will be executed by the shell on the
        blade under an SSH session by creating a subprocess.Popen()
        object. If additional keyword arguments are supplied, they may
        be used to override defaults set up by this function and
        passed to subprocess.Popen() or simply passed on to
        subprocess.Popen() as keyword arguments.

        If the 'logfiles' argument is provided, it contains a two
        element tuple telling run_command where to put standard output
        and standard error logging for the command respectively.
        Normally, these are specified as pathnames to log
        files. Either or both can also be a file object or None. If a
        file object is used, the output is written to the file. If
        None is used, the corresponding output is not redirected and
        the default Popen() behavior is used.

        If the 'blocking' option is False (default), run_command() will
        block waiting for the command to complete (or fail) and raise
        a ContextualError exception if it fails. If the 'blocking' option
        is False, run_command() will return immediately once the
        Popen() object is created and let the caller manage the
        sub-process.

        """

    @abstractmethod
    def __enter__(self):
        """Context entry handler to make BladeSSHConnection objects
        usable with the 'with ... as' construct. This returns the
        BladeSSHConnection for use in a context.

        """

    @abstractmethod
    def __exit__(
            self,
            exception_type=None,
            exception_value=None,
            traceback=None
    ):
        """Context exit handler to make BladeSSHConnection objects
        usable with the 'with ... as' construct. This cleans up all
        resources associated with the BladeSSHConnection on exit from
        the 'with' block context. Not called if the object is used
        outside a 'with' context.

        """


class BladeSSHConnectionSetBase(BladeConnectionSetBase, metaclass=ABCMeta):
    """A class to wrap multiple BladeSSHConnections and provide
    operations that run in parallel across multiple connections.

    """
    @abstractmethod
    def copy_to(
        self, source, destination,
        recurse=False, logname=None, blade_class=None
    ):
        """Copy the file at a path on the local machine ('source') to
        a path ('dest') on all of the selected blades (based on
        'blade_class'). If 'blade_class is not specified or None, copy
        the file to all connected blades. Wait until all copies
        complete or fail. If any of the copies fail, collect the
        errors they produce to raise a ContextualError exception
        describing the failures.

        If the 'recurse' option is True and the local file is a
        directory, the directory and all of its descendants will be
        copied.

        If the 'logname' argument is provided, use the string found
        there to compose paths to two files, one that will contain the
        standard output from the command and one that will contain the
        standard input. The paths to these files will be included in
        any error reporting from the operation.

        """

    @abstractmethod
    def run_command(self, cmd, logname=None, blade_class=None):
        """Using SSH, run the command in the string 'cmd'
        asynchronously on all connected blades filtered by
        'blade_class'. If 'blade_class' is unspecified or None, run on
        all connected blades. The string 'cmd' can be templated using
        Jinja templating to use any of the attributes of the
        underlying connection. In this case, the connection in which
        the command is being run will be used for the templating, so,
        for example, 'blade_hostname' will match the blade on which
        the command runs:

        - the blade class: 'blade_class'
        - the blade instance within its class: 'instance'
        - the blade hostname: 'blade_hostname'
        - the connection port on the blade: 'remote_port'
        - the local connection IP address: 'local_ip'
        - the local connection port: 'local_port'

        Wait until all commands complete or fail. If any of the
        commands fail, collect the errors they produce to raise a
        ContextualError exception describing the failures.

        If the 'logname' argument is provided, use the string found
        there to compose paths to two files, one that will contain the
        standard output from the command and one that will contain the
        standard input. The paths to these files will be included in
        any error reporting from the operation.

        """

    @abstractmethod
    def __enter__(self):
        """Context entry handler to make BladeSSHConnectionSet objects
        usable with the 'with ... as' construct. This returns the
        BladeSSHConnectionSet for use in a context.

        """

    @abstractmethod
    def __exit__(
            self,
            exception_type=None,
            exception_value=None,
            traceback=None
    ):
        """Context exit handler to make BladeSSHConnectionSet objects
        usable with the 'with ... as' construct. This cleans up all
        resources associated with the BladeSSHConnectionSet on exit
        from the 'with' block context. Not called if the object is
        used outside a 'with' context.

        """


class SecretsBase(metaclass=ABCMeta):
    """Provider Layers Secrets API object. Provides ways to populate
    and retrieve secrets through the Provider layer. Secrets are
    created by the provider layer by declaring them in the Provider
    configuration for your vTDS system, and should be known by their
    names as filled out in various places and verious layers in your
    vTDS system. For example the SSH key pair used to talk to a
    particular set of Virtual Blades through a blade connection is
    stored in a secret configured in the Provider layer and the name
    of that secret can be obtained from a VirtualBlades API object
    using the blade_ssh_key_secret() method.

    """
    @abstractmethod
    def store(self, name, value):
        """Store a value (string) in the named secret.

        """

    @abstractmethod
    def read(self, name):
        """Read the value (string) stored in a named secret. If no
        value is present, return None.

        """

    @abstractmethod
    def application_metadata(self, name):
        """Get the application metadata for a named secret from its
        config.

        """



#
# MIT License
#
# (C) Copyright [2024] Hewlett Packard Enterprise Development LP
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
"""Public API module for the kvm cluster layer, this gives callers
access to the Cluster API and prevents them from seeing the private
implementation of the API.

"""
from abc import (
    ABCMeta,
    abstractmethod
)
from ..layerapi import LayerAPI


class ClusterAPI(LayerAPI, metaclass=ABCMeta):
    """ Presents the Cluster API to callers.

    """
    @abstractmethod
    def get_virtual_nodes(self):
        """Return a VirtualNodes object containing all of the
        available non-pure-base-class Virtual Nodes.

        """

    @abstractmethod
    def get_virtual_networks(self):
        """Return a VirtualNetworks object containing all the
        available VirtualNetworks.

        """



#
# MIT License
#
# (C) Copyright [2024] Hewlett Packard Enterprise Development LP
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
"""The base layer API class which provides the methods (in pure
virtual form) used by vtds-core to manage vTDS systems. All layers
implement thes methods.

"""
from abc import (
    ABCMeta,
    abstractmethod
)


class LayerAPI(metaclass=ABCMeta):
    """A class defining the minimum Layer API for all vTDS layer
    implementations.

    """
    @abstractmethod
    def consolidate(self):
        """Consolidate layer configurations, carrying out any actions
        that might affect other layer configurations. At the end of
        this phase, the overall configuration is fully specified and
        ready for use by prepare and validate.

        """

    @abstractmethod
    def prepare(self):
        """Prepare data files and other state needed for
        deployment. At the end of this stage the layer is ready to be
        displayed, validated, deployed or removed.

        """

    @abstractmethod
    def validate(self):
        """Run any configuration validation that may be appropriate
        for the layer and fail if that validation
        fails. Prerequisites: consolidate() and prepare().

        """

    @abstractmethod
    def deploy(self):
        """Deploy the layer. Prerequisites: consolidate() and prepare().

        """

    @abstractmethod
    def remove(self):
        """Remove respources owned by the layer. Prerequisites:
        consolidate() and prepare().

        """



#
# MIT License
#
# (C) Copyright 2024 Hewlett Packard Enterprise Development LP
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
"""Entrypoint handling.

"""
import sys
from .errors import (
    ContextualError,
    UsageError,
    usage,
    error_msg
)


def entrypoint(usage_msg, main):
    """Generic entrypoint function. This sets up command line
    arguments for the invocation of a 'main' function and takes care
    of handling any vTDS exceptions that are raised to report
    errors. Other exceptions are allowed to pass to the caller for
    handling.

    """
    try:
        main(sys.argv[1:])
    except ContextualError as err:
        error_msg(str(err))
        sys.exit(1)
    except UsageError as err:
        usage(usage_msg, str(err))



#
# MIT License
#
# (C) Copyright 2024 Hewlett Packard Enterprise Development LP
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
"""Operations on configuration structures.

"""

from os.path import join as path_join
from .errors import ContextualError
from .config_operations import read_config


# pylint: disable=too-few-public-methods
class BaseConfiguration:
    """BaseConfig class presents operations on the base configuration
    of the provider layer to callers.

    """
    def __init__(
            self,
            description,
            config_dir,
            config_file="config.yaml",
            test_overlay="test_overlay.yaml"
    ):
        """Constructor

        """
        self.config_dir = config_dir
        self.config_file = config_file
        self.test_overlay = test_overlay
        self.description = description

    def get_base_config(self):
        """Retrieve the base configuration for the provider in the
        form of a python data structure for use in composing and
        overall vTDS configuration.

        """
        config = path_join(self.config_dir, self.config_file)
        description = self.description + " base configuration"
        return read_config(config, description)

    def get_test_overlay(self):
        """Retrieve a pre-defined test overlay configuration in the
        form of a python data structure for use in composing vTDS
        configurations for testing with this provider layer.

        """
        config = path_join(self.config_dir, self.test_overlay)
        description = self.description + " test configuration overlay"
        return read_config(config, description)

    def get_base_config_text(self):
        """Retrieve the text of the base configuration file as a text
        string (UTF-8 encoded) for use in displaying the configuration
        to users.

        """
        config = path_join(self.config_dir, self.config_file)
        try:
            with open(config, 'r', encoding='UTF-8') as config_stream:
                return config_stream.read()
        except OSError as err:
            raise ContextualError(
                "cannot open %s base config file '%s' - %s" % (
                    self.description,
                    config, str(err)
                )
            ) from err



#
# MIT License
#
# (C) Copyright [2024] Hewlett Packard Enterprise Development LP
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
"""Objects presented on the Layer API containing public information
and operations in the provider layer.

"""
from contextlib import contextmanager
from abc import (
    ABCMeta,
    abstractmethod
)


class VirtualNodesBase(metaclass=ABCMeta):
    """A class implementing public access to Virtual Nodes and their
    operations.

    """
    @abstractmethod
    def node_classes(self):
        """Get a list of Virtual Node classes by name.

        """

    @abstractmethod
    def application_metadata(self, node_class):
        """Get the application metadata for a named node class from
        its config.

        """

    @abstractmethod
    def node_count(self, node_class):
        """Get the number of Virtual Node instances of the specified
        class.

        """

    @abstractmethod
    def set_node_node_name(self, node_class, instance, name):
        """When called during the 'prepare' phase by a higher layer,
        this sets the contextual node name (i.e. the node name used on
        the mechanism used to implement the Virtual Node, perhaps a
        Virtual Machine) to be used for a given node instance in a
        given node class to the specified name. This modifies the
        configuration data used during the deploy and remove phases.

        """

    @abstractmethod
    def node_node_name(self, node_class, instance):
        """Get the node name (used for contextually naming the
        implementation of a Virtual Node) of a given Virtual Node
        class instance (i.e. Virtual Node). This will also be the core
        hostname (minus the network name) of the node if there is no
        host naming specified in the configuration.

        """

    @abstractmethod
    def network_names(self, node_class):
        """Return the list of Network Names of the Networks connected
        to the specified class of Virtual Node.

        """

    @abstractmethod
    def set_node_hostname(self, node_class, instance, name):
        """When called during the 'prepare' phase by a higher layer,
        this sets the network hostname to be used for a given node
        instance in a given node class to the specified name. This
        modifies the configuration data used during the deploy and
        remove phases.

        """

    @abstractmethod
    def node_hostname(self, node_class, instance, network_name=None):
        """Get the hostname of a given Virtual Node class instance
        (i.e. Virtual Node). Optionally, look up the hostname by
        network name for local names on a given Virtual Network.

        """

    def node_host_blade_info(self, node_class):
        """Get the host Virtual Blade class information from a a given
        Virtual Node class instance (i.e. Virtual Node). This returns
        a dictionary with the following elements:

        - blade_class: the name of the Virtual Blade class to which
          the blade belongs.

        - instance_capacity: the maximum number of host instances that
          will be created per blade.

        """

    @abstractmethod
    def node_ipv4_addr(self, node_class, instance, network_name):
        """Get the configured IPv4 address (if any) for the specified
        instance of the specified node class on the specified
        network. If IP addresses are not configured for the specified
        node class (i.e. they are dynamic DHCP addresses) this will
        return None. If the specified node class is not present on the
        specified network this will raise a ContextualError exception.

        """

    @abstractmethod
    def node_class_addressing(self, node_class, network_name):
        """Obtain an Addressing object describing the addressing of
        the specified node class within the named network. If the node
        class has no addressing in (i.e. is not connected to) the
        specified network, return None.

        """

    @abstractmethod
    def node_ssh_key_secret(self, node_class):
        """Return the name of the secret containing the SSH key pair
        used to to authenticate with Virtual Node of the specified
        node class.

        """

    @abstractmethod
    def node_ssh_key_paths(self, node_class):
        """Return a tuple of paths to files containing the public and
        private SSH keys used to to authenticate with Virtual Nodes of
        the specified node class. The tuple is in the form
        '(public_path, private_path)' The value of 'private_path' is
        suitable for use with the '-i' option of 'ssh'. Before
        returning this call will verify that both files can be opened
        for reading and will fail with a ContextualError if either
        cannot.

        """

    @abstractmethod
    @contextmanager
    def connect_node(self, node_class, instance, remote_port):
        """Establish an external connection to the specified remote
        port on the specified instance of the named Virtual Node
        class. Return a context manager (suitable for use in a 'with'
        clause) yielding an NodeConnection object for the
        connection. Upon leaving the 'with' context, the connection in
        the NodeConnection is closed.

        """

    @contextmanager
    @abstractmethod
    def connect_nodes(self, remote_port, node_classes=None):
        """Establish external connections to the specified remote port
        on all the Virtual Node instances on all the Virtual Node
        classes listed by name in 'node_classes'. If 'node_classes' is not
        provided or None, all available node classes are used. Return a
        context manager (suitable for use in a 'with' clause) yielding
        a NodeConnectionSet object representing the connections. Upon
        leaving the 'with' context, all the connections in the
        NodeConnectionSet object are closed.

        """

    @abstractmethod
    @contextmanager
    def ssh_connect_node(self, node_class, instance, remote_port=22):
        """Establish an external connection to the SSH server
        port on the specified instance of the named Virtual Node
        class. Return a context manager (suitable for use in a 'with'
        clause) yielding a NodeSSHConnection object for the
        connection. Upon leaving the 'with' context, the connection in
        the NodeSSHConnection is closed.

        """

    @contextmanager
    @abstractmethod
    def ssh_connect_nodes(self, node_classes=None, remote_port=22):
        """Establish external connections to the specified remote port
        on all the Virtual Node instances on all the Virtual Node
        classes listed by name in 'node_classes'. If 'node_classes' is not
        provided or None, all available node classes are used. Return a
        context manager (suitable for use in a 'with' clause) yielding
        a NodeSSHConnectionSet object representing the connections. Upon
        leaving the 'with' context, all the connections in the
        NodeSSHConnectionSet are closed.

        """


class VirtualNetworksBase(metaclass=ABCMeta):
    """The external representation of the set of Virtual Networks
    and public operations that can be performed on the list..

    """
    @abstractmethod
    def network_names(self):
        """Get a list of network names

        """

    @abstractmethod
    def application_metadata(self, network_name):
        """Get the application metadata for a named virtual network
        from its config.

        """

    @abstractmethod
    def ipv4_cidr(self, network_name):
        """Return the (string) IPv4 CIDR (<IP>/<length>) for the
        named network.

        """

    @abstractmethod
    def non_cluster_network(self, network_name):
        """A network in the cluster configuration may be a network
        used by vTDS for constructing the cluster but not intended for
        use by applications running on the cluster. Such a network is
        a 'non-cluster network'.  Return True if the specified network
        name refers to a non-cluster network otherwise return False.

        """

    @abstractmethod
    def blade_class_addressing(self, blade_class, network_name):
        """Obtain an Addressing object describing the addressing of
        the specified connected blade class within the named
        network. If the blade class has no addressing in (i.e. is not
        connected to) the specified network, return None.

        """


class NodeConnectionBase(metaclass=ABCMeta):
    """A class containing the relevant information needed to use
    external connections to ports on a specific Virtual Node.

    """
    @abstractmethod
    def node_class(self):
        """Return the name of the Virtual Node class of the connected
        Virtual Node.

        """

    @abstractmethod
    def node_hostname(self, network_name=None):
        """Return the hostname of the connected Virtual Node. If
        network_name is specified and not none, return the local
        hostname on the specified Virtual Network.

        """

    @abstractmethod
    def local_ip(self):
        """Return the locally reachable IP address of the connection
        to the Virtual Node.

        """

    @abstractmethod
    def local_port(self):
        """Return the TCP port number on the locally reachable IP
        address of the connection to the Virtual Node.

        """

    @abstractmethod
    def remote_port(self):
        """Return the TCP port number on the on the Virtual Node to
        which the NodeConnection connects.

        """


class NodeConnectionSetBase(metaclass=ABCMeta):
    """A class that contains multiple active NodeConnections to
    facilitate operations on multiple simultaneous nodes. This class
    is just a wrapper for a list of NodeConnections and should be
    obtained using the VirtualNodes.connect_nodes() method not
    directly.

    """
    @abstractmethod
    def list_connections(self, node_class=None):
        """List the connections in the NodeConnectionSet filtered by
        'node_class' if that is present. Otherwise simply list all of
        the connections.

        """

    @abstractmethod
    def get_connection(self, hostname):
        """Return the connection corresponding to the specified
        VirtualNode hostname ('hostname') or None if the hostname is
        not found.

        """


class NodeSSHConnectionBase(NodeConnectionBase, metaclass=ABCMeta):
    """Specifically a connection to the SSH server on a node (remote
    port 22 unless otherwise specified) with methods to copy files to
    and from the node using SCP and to run commands on the node
    using SSH.

    """
    @abstractmethod
    def copy_to(
        self, source, destination,
        recurse=False, blocking=True, logname=None, **kwargs
    ):
        """Copy a file from a path on the local machine ('source') to
        a path on the Virtual Node ('dest'). The SCP operation is run
        under a subprocess.Popen() object, which is returned at the
        end of the call. If additional keyword arguments are supplied,
        they may be used to override defaults set up by this function
        and passed to subprocess.Popen() or simply passed on to
        subprocess.Popen() as keyword arguments.

        If the 'recurse' argument is 'True' and the source is a
        directory, the directory and all of its descendents will be
        copied. Otherwise, the source should be a file and it alone
        will be copied.

        If the 'blocking' option is True (default), copy_to() will block
        waiting for the copy to complete (or fail) and raise a
        ContextualError exception if it fails. If the 'blocking' option
        is False, copy_to() will return immediately once the Popen()
        object is created and let the caller manage the sub-process.

        If the 'logname' argument is provided and not None, use the
        string found there to compose a pair of log files to capture
        standard output and standard error. Otherwise a generic log
        name is created.

        """

    @abstractmethod
    def copy_from(
        self, source, destination,
        recurse=False, blocking=True, logname=None, **kwargs
    ):
        """Copy a file from a path on the node ('source') to a path
        on the local machine ('dest'). The SCP operation is run under
        a subprocess.Popen() object, which is returned at the end of
        the call. If additional keyword arguments are supplied, they
        may be used to override defaults set up by this function and
        passed to subprocess.Popen() or simply passed on to
        subprocess.Popen() as keyword arguments.

        If the 'recurse' argument is 'True' and the source is a
        directory, the directory and all of its descendents will be
        copied. Otherwise, the source should be a file and it alone
        will be copied.

        If the 'blocking' option is True (default), copy_from() will
        block waiting for the copy to complete (or fail) and raise a
        ContextualError exception if it fails. If the 'blocking' option
        is False, copy_from() will return immediately once the Popen()
        object is created and let the caller manage the sub-process.

        If the 'logname' argument is provided and not None, use the
        string found there to compose a pair of log files to capture
        standard output and standard error. Otherwise a generic log
        name is created.

        """

    @abstractmethod
    def run_command(self, cmd, blocking=True, logfiles=None, **kwargs):
        """Using SSH, run the command in the string 'cmd' on the
        node. The string 'cmd' can be templated using Jinja
        templating to use any of the attributes of the underlying
        connection:

        - the node class: 'node_class'
        - the node hostname: 'node_hostname'
        - the connection port on the node: 'remote_port'
        - the local connection IP address: 'local_ip'
        - the local connection port: 'local_port'

        The resulting command will be executed by the shell on the
        node under an SSH session by creating a subprocess.Popen()
        object. If additional keyword arguments are supplied, they may
        be used to override defaults set up by this function and
        passed to subprocess.Popen() or simply passed on to
        subprocess.Popen() as keyword arguments.

        If the 'logfiles' argument is provided, it contains a two
        element tuple telling run_command where to put standard output
        and standard error logging for the command respectively.
        Normally, these are specified as pathnames to log
        files. Either or both can also be a file object or None. If a
        file object is used, the output is written to the file. If
        None is used, the corresponding output is not redirected and
        the default Popen() behavior is used.

        If the 'blocking' option is False (default), run_command() will
        block waiting for the command to complete (or fail) and raise
        a ContextualError exception if it fails. If the 'blocking' option
        is False, run_command() will return immediately once the
        Popen() object is created and let the caller manage the
        sub-process.

        """


class NodeSSHConnectionSetBase(NodeConnectionSetBase, metaclass=ABCMeta):
    """A class to wrap multiple NodeSSHConnections and provide
    operations that run in parallel across multiple connections.

    """
    @abstractmethod
    def copy_to(
        self, source, destination, recurse=False, logname=None, node_class=None
    ):
        """Copy the file at a path on the local machine ('source') to
        a path ('dest') on all of the selected nodes (based on
        'node_class'). If 'node_class is not specified or None, copy
        the file to all connected nodes. Wait until all copies
        complete or fail. If any of the copies fail, collect the
        errors they produce to raise a ContextualError exception
        describing the failures.

        If the 'recurse' option is True and the local file is a
        directory, the directory and all of its descendants will be
        copied.

        If the 'logname' argument is provided, use the string found
        there to compose paths to two files, one that will contain the
        standard output from the command and one that will contain the
        standard input. The paths to these files will be included in
        any error reporting from the operation.

        """

    @abstractmethod
    def run_command(self, cmd, logname=None, node_class=None):
        """Using SSH, run the command in the string 'cmd'
        asynchronously on all connected nodes filtered by
        'node_class'. If 'node_class' is unspecified or None, run on
        all connected nodes. The string 'cmd' can be templated using
        Jinja templating to use any of the attributes of the
        underlying connection. In this case, the connection in which
        the command is being run will be used for the templating, so,
        for example, 'node_hostname' will match the node on which
        the command runs:

        - the node class: 'node_class'
        - the node instance within its class: 'instance'
        - the node hostname: 'node_hostname'
        - the connection port on the node: 'remote_port'
        - the local connection IP address: 'local_ip'
        - the local connection port: 'local_port'

        Wait until all commands complete or fail. If any of the
        commands fail, collect the errors they produce to raise a
        ContextualError exception describing the failures.

        If the 'logname' argument is provided, use the string found
        there to compose paths to two files, one that will contain the
        standard output from the command and one that will contain the
        standard input. The paths to these files will be included in
        any error reporting from the operation.

        """


class AddressingBase(metaclass=ABCMeta):
    """Addressing information for node and blade classes. This
    contains all addressing by address family for instances of node
    class or blade classes as assigned at the cluster level.

    """
    @abstractmethod
    def address(self, family, instance):
        """Retrieve the address within the address family specified by
        the string 'family' (AF_INET, AF_INET6, AF_PACKET, or
        whatever) of a specified instance number (integer) of the
        object class (node class or blade class) and network from
        which this Addressing object was obtained. The form of the
        address will be a string formatted in the canonical way for
        the specified address family.

        """

    @abstractmethod
    def addresses(self, family):
        """Retrieve a dictionary of instance number (integer) to
        address (string) mapping pairs within the address family
        specified by the string 'family' (AF_INET, AF_INET6,
        AF_PACKET, or whatever) of the object class (node class or
        blade class) and network from which this Addressing object was
        obtained.

        """

    @abstractmethod
    def address_families(self):
        """Retrieve a list of address family names ('AF_INET',
        'AF_INET6', 'AF_PACKET', or whatever) supported by the object
        class (node class or blade class) and network from which this
        Addressing object was obtained.

        """

    @abstractmethod
    def instances(self):
        """Retrieve a list of instance numbers within the object class
        (node class or blade class) and network from which this
        addressing object was created. Only instances that are
        connected to the network are reported.

        """



#
# MIT License
#
# (C) Copyright 2024 Hewlett Packard Enterprise Development LP
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
"""Logfile related tools

"""
from os.path import join as path_join
from os import (
    makedirs,
    devnull
)
from io import IOBase
from contextlib import contextmanager
from .errors import ContextualError


@contextmanager
def logfile(lfile, mode='w', encoding='UTF-8',  **kwargs):
    """Open a logfile and yield the resulting stream if 'lfile' is not
       None and is a string. If 'lfile' is None, yield a writable NULL
       stream (devnull). If 'lfile' is a stream (IOBase derived),
       yield 'lfile' itself and make sure not to close it on context
       exit.

    """
    stream = None
    try:
        if lfile is None:
            lfile = devnull
        if isinstance(lfile, str):
            stream = open(
                lfile, mode, encoding=encoding, **kwargs
            ) if lfile is not None else None
        elif isinstance(lfile, IOBase):
            stream = lfile
        else:
            raise ContextualError(
                "error opening a log file, expected the specified "
                "file to be a string, a stream or None, "
                "not '%s'" % str(type(lfile))
            )
    except OSError as err:
        raise ContextualError(
            "error opening log file '%s' - %s" % (lfile, str(err))
        ) from err
    try:
        yield stream
    finally:
        if isinstance(lfile, str) and stream is not None:
            stream.close()


def log_paths(build_dir, logname):
    """Compose an 'out' path and an 'error' path based on the base
    name 'logname' and return both to the caller.

    """
    logs = path_join(build_dir, "logs")
    try:
        makedirs(logs, mode=0o755, exist_ok=True)
    except OSError as err:
        raise ContextualError(
            "failed to create log directory '%s' - %s" % (
                logs, str(err)
            )
        ) from err
    out_path = path_join(
        logs,
        "%s-out.txt" % (logname)
    )
    err_path = path_join(
        logs,
        "%s-err.txt" % (logname)
    )
    return (out_path, err_path)



#
# MIT License
#
# (C) Copyright [2024] Hewlett Packard Enterprise Development LP
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
""" Nox definitations for tests, docs, and linting
"""
from __future__ import absolute_import
import os

import nox  # pylint: disable=import-error


COVERAGE_FAIL = 95

PYTHON = False if os.getenv("NOX_DOCKER_BUILD") else ['3']

@nox.session(python=PYTHON)
def lint(session):
    """Run linters.
    Returns a failure if the linters find linting errors or sufficiently
    serious code quality issues.
    """
    run_cmd = ['pylint', 'vtds_base']
    if session.python:
        session.install('.[lint]')
    session.run(*run_cmd)


@nox.session(python=PYTHON)
def style(session):
    """Run code style checker.
    Returns a failure if the style checker fails.
    """
    run_cmd = ['pycodestyle',
               '--config=.pycodestyle',
               'vtds_base']
 
    if session.python:
        session.install('.[style]')
    session.run(*run_cmd)


@nox.session(python=PYTHON)
def test(session):
    """Default unit test session.
    """
    # Install all test dependencies, then install this package in-place.
    path = 'tests'
    if session.python:
        session.install('.[test]')

    # Run py.test against the tests. XXX - Disabled until we have unit tests
#    session.run(
#        'py.test',
#        '--quiet',
#        '-W',
#        'ignore::DeprecationWarning',
#        '--cov=vtds_base',
#        '--cov=tests',
#        '--cov-append',
#        '--cov-config=.coveragerc',
#        '--cov-report=',
#        '--cov-fail-under={}'.format(COVERAGE_FAIL),
#        os.path.join(path),
#        env={}
#    )
#    session.run(
#        'coverage', 'report',
#        '--show-missing',
#        '--fail-under={}'.format(COVERAGE_FAIL)
#    )
#    session.run('coverage', 'erase')
    session.run('true') # Do something, but not much for now...
